/***
*** pci16AI64_MAIN.C 
***
***  General description of this file:
***  	Device driver source code for General Standards PCI-16AI64
***  	64 channel PCI A/D board. This file is part of the Linux
***  	driver source distribution for this board.
***  	
***  	This file is not necessary to compile application programs, therefore 
***  	should NOT be included in binary only driver distributions.
***
***  Copyrights (c):
***  	General Standards Corporation (GSC), 2002-2003
***
***  Author:
***  	Evan Hillman, GSC Inc. (evan@generalStandards.com)
***
***  Support:
***  	Primary support for this driver is provided by GSC. 
***
***  Platform (tested on, may work with others):
***  	Linux, kernel version 2.4.x, Intel hardware.
***
***/

#ifndef __KERNEL__
#define __KERNEL__
#endif

#ifndef MODULE
#define MODULE
#endif

#include <linux/module.h>
#include <linux/fs.h>
#include <linux/vmalloc.h>
#include <linux/pci.h>
#include <linux/mm.h>
#include <linux/slab.h>
#include <linux/sched.h>
#include <linux/tqueue.h>
#include <linux/interrupt.h>
#include <linux/timer.h>
#include <linux/ioctl.h>
#include <linux/proc_fs.h>

#include <asm/io.h>
#include <asm/system.h>
#include <asm/uaccess.h>

#include "sysdep.h"
#include "gsc16AI64_ioctl.h"
#include "gsc16AI64.h"
#include "plx_regs.h"

MODULE_LICENSE("GPL");

struct board_entry boards_supported[] = {
	/*    subvendor    subsystem name       index ID */
	{GSC_SUBVENDOR,0x2406,"PMC-12AI64",gsc12ai64}, /* PMC-12AI64 */
	{GSC_SUBVENDOR,0x2407,"PMC-16AI64",gsc16ai64}, /* PMC-16AI64 */
	{GSC_SUBVENDOR,0x2868,"PMC-16AI64SS",gsc16ai64ss}, /** PCM-16AI64SS */
	{0,0,"null"}               /* null terminated list.... */
};

/* Driver Version */
static const char device_version[] = "source version - " DRIVER_VERSION " built - " "$Date: " __DATE__ " " __TIME__ " $";

/* module load/unload functions */
int init_module(void);
void cleanup_module(void);

/* local functions */
static int device_write(struct file *, const char *, size_t, loff_t *);
static int device_open(struct inode *, struct file *);
static int device_close(struct inode *, struct file *);

/* local module data */
static struct file_operations device_fops = {
open:    device_open,
release: device_close,
read:    device_read,
write:   device_write,
ioctl:   device_ioctl
};

static	__s8	version[8];
static	__s8	built[32];

static int proc_enabled;
static int num_boards;
static int device_major = 0;     	      /* device major number (dynamically assigned) */
static struct device_board *boards = 0;  	  /* linked list of all pci16AI64 boards found */

#ifdef DEBUG
ulong int_other_count[MAX_BOARDS];
ulong int_count[MAX_BOARDS];
ulong dma_count[MAX_BOARDS];
ulong channel_irq[MAX_BOARDS];
ulong channel_expected[MAX_BOARDS];
int board_type[MAX_BOARDS];
void * context[MAX_BOARDS];

/************************************************************************/
/* regdump - debug only                                                 */
/************************************************************************/
void regdump(struct device_board *device)
{
	int q;

	for (q=0;q<8;q++)
	{
		printk(KERN_INFO GSC_NAME ": register %d value: %.8x\n", q, readlocal(device,q));
	}
	printk(KERN_INFO GSC_NAME " PLX IRQ %.8X.\n",readl(IntCntrlStat(device)));
}
#endif

/************************************************************************/
/* timeout_handler                                                      */
/************************************************************************/
void timeout_handler(unsigned long ptr)
{
	struct device_board *device = (struct device_board *) ptr;
	unsigned long regval;
#if DEBUG
	writel(0x22222222, (Mailbox0(device)));
#endif
	printk(KERN_ERR GSC_NAME "(%d): ERROR - device timeout\n", device->minor);
	device->timeout=TRUE;
#ifdef DEBUG
	{
		long fifosamples,threshold;
		fifosamples = readlocal(device,BUFFER_SIZE_REG);
		threshold = readlocal(device, INPUT_BUFF_CTRL_REG)&IDC_INPUT_BUFFER_SIZE;
		printk(KERN_INFO GSC_NAME "(%d): timeout -> fifo samples %.8lX threshold %.8lX\n", device->minor, fifosamples, threshold);
	}
	printk("timeout -> plx int %.8X board control %.8X local int %.8X DMA status %.8X buffer size: %.8X \n",	
		readl(IntCntrlStat(device)), readlocal(device,BOARD_CTRL_REG),readlocal(device,INT_CTRL_REG),
		readl(DMACmdStatus(device)),readlocal(device,BUFFER_SIZE_REG));
	printk("DMA Mode %.8X\n",readl(DMAMode0(device)));
	printk("Firm rev %.8X\n",readlocal(device,0x3c/4));

	printk(KERN_INFO GSC_NAME "(%d):    init_pending %d\n", device->minor, device->init_pending);
	printk(KERN_INFO GSC_NAME "(%d): autocal_pending %d\n", device->minor, device->autocal_pending);
	printk(KERN_INFO GSC_NAME "(%d):   lo_hi_pending %d\n", device->minor, device->lo_hi_pending);
	printk(KERN_INFO GSC_NAME "(%d): chready_pending %d\n", device->minor, device->chready_pending);
	printk(KERN_INFO GSC_NAME "(%d):   dma_pending %d\n", device->minor, device->dma_pending);

#endif
	writel(readl(IntCntrlStat(device)) & ~(IRQ_PCI_ENABLE | IRQ_LOCAL_PCI_ENABLE), IntCntrlStat(device));

	/* wake up whatever was waiting. */

	if (device->init_pending)
	{
		device->init_pending = FALSE;
		wake_up(&device->ioctlwq);
	}

	if (device->autocal_pending)
	{
		device->autocal_pending = FALSE;
		wake_up(&device->ioctlwq);
	}

	if (device->chready_pending)
	{ 
		device->chready_pending = FALSE;
		wake_up(&device->dmawq);
	}

	if (device->lo_hi_pending)
	{ 
		device->lo_hi_pending = FALSE;
		wake_up(&device->dmawq);
	}

	if (device->dma_pending)
	{ 
		regval = readl(DMACmdStatus(device));
		regval &=~(DMA0_ENABLE);
		writel(regval, DMACmdStatus(device));
		regval = DMA0_ABORT;
		writel(regval, DMACmdStatus(device));

		device->dma_pending = FALSE;
		wake_up(&device->dmawq);
	}

	// disable the PCI interrupt.
	writelocal(device,0, INT_CTRL_REG);
#if DEBUG
	printk(KERN_INFO GSC_NAME "(%d): timeout return\n", device->minor);
#endif
}

/******************************************************************************
*
*	Function:	read_proc
*
*	Purpose:
*
*		Implement the read service for the module's /proc file system
*		entry. Read the documentation on this service for details, as
*		we ignore some arguments according to our needs and the
*		documentation.
*
*	Arguments:
*
*		page	The data produced is put here.
*
*		start	Records pointer to where the data in "page" begins.
*
*		offset	The offset into the file where we're to begin.
*
*		count	The limit to the amount of data we can produce.
*
*		eof	Set this flag when we hit EOF.
*
*		data	A private data item that we may use, but don't.
*
*	Returned:
*
*		int	The number of characters written.
*
******************************************************************************/
static int read_proc(
					 char*	page,
					 char**	start,
					 off_t	offset,
					 int	count,
					 int*	eof,
					 void*	data)
{
#define	_PROC_MSG_SIZE	128

#if PAGE_SIZE < _PROC_MSG_SIZE
#error	"/proc file size may be too big."
#endif
	int	i;
#ifdef DEBUG
	int j;
	char	str[256];
#endif

	MOD_INC_USE_COUNT;
	i	= sprintf(page,
		"version: %s\n"
		"built: %s\n"
		"boards: %d\n",
		DRIVER_VERSION,
		built,
		num_boards);
	page[i]	= 0;

#ifdef DEBUG
	strcat(page, "Debug Stats:\n ");
	strcat(page,"type\tother_ints\ttotal_ints\tdma_ints\tchan_irq\tchan_expect\n");

	for(j=0;j<num_boards;j++){
		sprintf(str,"%d\t\t%ld\t\t%ld\t\t%ld\t\t%ld\t\t%ld\t\t%p\n",
			board_type[j],
			int_other_count[j],
			int_count[j],
			dma_count[j],
			channel_irq[j],
			channel_expected[j],
			context[j]
			);
			strcat(page,str);
	}

#endif
	strcat(page, "\n");
	i	= strlen(page) + 1;

	if (i >= PAGE_SIZE) {
		printk(KERN_ERR GSC_NAME "/proc/xxx is larger than PAGE_SIZE\n");
		i= PAGE_SIZE-1;
	}

	i--;
	eof[0]	= 1;
	MOD_DEC_USE_COUNT;
	return(i);
}

/******************************************************************************
*
*	Function:	proc_get_info
*
*	Purpose:
*
*		Implement the get_info() service for /proc file system support.
*
*	Arguments:
*
*		page	The data produced is put here.
*
*		start	Records pointer to where the data in "page" begins.
*
*		offset	The offset into the file where we're to begin.
*
*		count	The limit to the amount of data we can produce.
*
*		dummy	A parameter that is unused (for kernel 2.2 only).
*
*	Returned:
*
*		int	The number of characters written.
*
******************************************************************************/
int proc_get_info(
				  char*	page,
				  char**	start,
				  off_t	offset,
				  int	count,
				  int dummy)
{
	int	eof;
	int	i;

	i	= read_proc(page, start, offset, count, &eof, NULL);
	return(i);
}

/************************************************************************/
/* module initalization: detect card(s) and set up descriptor(s)        */
/************************************************************************/
int init_module(void)
{
	struct proc_dir_entry*	proc;
	unsigned short reg;
	struct device_board *device, *device_next;
	struct pci_dev *pdev;
	int index = 0;
	int i;
	int found;

	int ResourceCount;

	printk(KERN_INFO GSC_NAME ": driver (%s) version loading\n", device_version);
	printk(KERN_INFO GSC_NAME ": Copyright (C) 2002-2004 General Standards Corp.\n");

	strcpy(version, DRIVER_VERSION);
	sprintf(built, "%s, %s", __DATE__, __TIME__);

#ifdef LINUX_2_4
	pci_for_each_dev(pdev)
#else
	for (pdev = pci_devices; pdev; pdev = pdev->next)
#endif
	{
		//printk(KERN_INFO GSC_NAME ": Checking...vendor: %X device: %X subvendor: %X subsystem: %X\n",pdev->vendor, pdev->device, pdev->subsystem_vendor, pdev->subsystem_device);

		if ((pdev->vendor == PCI_VENDOR_ID_PLX) && (pdev->device == PCI_DEVICE_ID_PLX_9080)) {

			/* determine if this is one of the boards supported. */

			found=FALSE;
			i=0;
			while(boards_supported[i].subsystem_device != 0){
				if ((boards_supported[i].subsystem_device == pdev->subsystem_device) &&
					(boards_supported[i].subsystem_vendor == pdev->subsystem_vendor)){
						found=TRUE;
						//printk(KERN_INFO GSC_NAME ":     found board %s type %d\n",boards_supported[i].name,i);
					}
					if (found) break;
					i++;
			}

			if (found){
				pci_read_config_word(pdev, 0x2E, &reg);
				//printk(KERN_INFO GSC_NAME ": config reg=0x%x\n", reg);
				device = (struct device_board *)kmalloc(sizeof(struct device_board), GFP_KERNEL);
				printk(KERN_INFO GSC_NAME ": attaching board #%d device pointer %p\n", index + 1,device);
				device->board_type = i;
				/***/

				/*
				* The bars may be either in I/O space or memory space.  Figure out which
				* this board is and Deal with either.
				*/
				ResourceCount = 0;

				for (i = 0; i < PCI_NUM_BARS; ++i)
				{
					// Verify the address is valid
					if (pci_resource_start(
						pdev,
						i
						) == 0)
					{
						continue;
					}

					//printk(KERN_INFO GSC_NAME "   Resource %02d\n", ResourceCount);

					// Increment resource count
					ResourceCount++;
					// Get PCI physical address
					device->PciBar[i].Physical.QuadPart =
						pci_resource_start(
						pdev,
						i
						);

					// Determine resource type
					if (pci_resource_flags(
						pdev,
						i
						) & IORESOURCE_IO)
					{
						//printk(KERN_INFO GSC_NAME "     Type     : I/O Port\n");

						// Make sure flags are cleared properly
						device->PciBar[i].Physical.QuadPart &= ~(0x3);
						device->PciBar[i].IsIoMapped         = TRUE;
					}
					else
					{
						//printk(KERN_INFO GSC_NAME "     Type     : Memory Space\n");

						// Make sure flags are cleared properly
						device->PciBar[i].Physical.QuadPart &= ~(0xf);
						device->PciBar[i].IsIoMapped         = FALSE;
					}

					//printk(KERN_INFO GSC_NAME "     Address  : %08x\n",device->PciBar[i].Physical.u.LowPart);

					// Get the size - fake it for now, since it is small and fixed.
					device->PciBar[i].Size = 4096; /*
												   Plx_pci_resource_len(
												   device,
												   i
												   );*/
					/*
					if (device->PciBar[i].Size >= (1 << 10))
					{
					printk (KERN_INFO GSC_NAME"     Size     : %08x  (%d Kb)\n",	device->PciBar[i].Size, device->PciBar[i].Size >> 10);
					}
					else
					{
					printk(KERN_INFO GSC_NAME"     Size     : %08x  (%d bytes)\n",device->PciBar[i].Size, device->PciBar[i].Size);
					}
					*/
				}

#ifdef LINUX_2_4
				if (device->PciBar[2].IsIoMapped)
				{
					printk(KERN_INFO GSC_NAME ": I/O mapped local registers\n");
					device->local_addr = (u32 *)pdev->resource[2].start;
				}
				else
				{
					printk(KERN_INFO GSC_NAME ": memory mapped local registers\n");
					device->local_addr = (u32 *)ioremap(pdev->resource[2].start, 4096);
				}
				device->runtime_addr = (u32 *)ioremap(pdev->resource[0].start, 4096);

				init_waitqueue_head(&device->ioctlwq);
				init_waitqueue_head(&device->readwq);
				init_waitqueue_head(&device->dmawq);
				printk(KERN_INFO GSC_NAME ": base_address[2]=0x%lx, base_address[0]=0x%lx\n", pdev->resource[2].start, pdev->resource[0].start);
#else
				if (device->PciBar[2].IsIoMapped)
				{
					printk(KERN_INFO GSC_NAME ": I/O mapped local registers\n");
					device->local_addr = base_address[2];
				}
				else
				{
					printk(KERN_INFO GSC_NAME ": memory mapped local registers\n");
					device->local_addr = (u32 *)ioremap(pdev->resource[2].start, 4096);
				}
				device->runtime_addr = (u32 *)ioremap(pdev->base_address[0], 4096);
				device->ioctlwq = NULL;
				device->readwq = NULL;
				device->dmawq = NULL;
				printk(KERN_INFO GSC_NAME ": base_address[2]=0x%lx, base_address[0]=0x%lx\n", pdev->base_address[2], pdev->base_address[0]);
#endif
				/***/
				printk(KERN_INFO GSC_NAME ": runtime addr=%p, local addr=%p\n", device->local_addr, device->runtime_addr);
				if (device->local_addr == 0 || device->runtime_addr == 0)
				{
					printk(KERN_ERR GSC_NAME ": ERROR - Unable to map memory region\n");
					return (-ENODEV);
				}

				device->pdev = pdev;
				device->irqlevel = pdev->irq;
				device->busy = 0;
				device->detectOverflow=FALSE;
				device->minor = index;
				device->next = boards;
#ifdef DEBUG
				//regdump(device);
#endif
				//printk(KERN_INFO GSC_NAME ": local_addr=%p, runtime_addr=%p irq=%d\n", device->local_addr, device->runtime_addr, device->irqlevel);
				boards = device;
				index++;
				num_boards++;
#ifdef DEBUG
				printk(KERN_INFO GSC_NAME ":     base_address[0]=0x%lX, base_address[2]=0x%lX\n",device->pdev->resource[0].start, device->pdev->resource[2].start);
				printk("Firmware revision: %.8X\n",readlocal(device,FIRMWARE_REV_REG));
#endif
			}
		}
	}
	if (index == 0) {
		printk(KERN_INFO GSC_NAME ": no board found\n");
		return (-ENODEV);
	}
	device_major = register_chrdev(0, GSC_NAME, &device_fops);
	if (device_major < 0) {
		/* OK, could not register -- undo the work we have done above */
		printk(KERN_INFO GSC_NAME ": could not register device number\n");
		for (device = boards; device; device = device_next) {
			device_next = device->next;
			iounmap(device->local_addr);
			iounmap(device->runtime_addr);
			kfree(device);
		}
		boards = NULL;
		return (-ENODEV);
	}
	printk(KERN_INFO GSC_NAME ": Number of devices %d\n", num_boards);
	/*
	*	Add /proc file system support.
	*/

	if (num_boards)
	{
		remove_proc_entry(GSC_NAME, NULL);	/* Just in case. */
		proc	= create_proc_entry(GSC_NAME, S_IRUGO, NULL);

		if (proc)
		{
			proc_enabled	= 1;
			proc->read_proc	= read_proc;
			proc->get_info	= (void*) proc_get_info;
		}
		else
		{
			printk(KERN_INFO GSC_NAME "<1>%s: create_proc_entry() failure.\n",
				GSC_NAME);
			cleanup_module();
		}
	}

	printk(KERN_INFO GSC_NAME ": major=%d\n", device_major);
	return 0;
}
/************************************************************************/
/* cleanup when unloading module                                        */
/************************************************************************/
void cleanup_module(void)
{
	struct device_board *device, *device_next;
	unregister_chrdev(device_major, GSC_NAME);
	for (device = boards; device; device = device_next) {
		device_next = device->next;
		iounmap(device->local_addr);
		iounmap(device->runtime_addr);
		kfree(device);
	}
	boards = NULL;
	if (proc_enabled)	{
		remove_proc_entry(GSC_NAME, NULL);
		proc_enabled	= 0;
	}

	printk(KERN_INFO GSC_NAME ": unloaded\n");
}

/************************************************************************/
/* open device                                                          */
/************************************************************************/
static int device_open(struct inode *inode, struct file *fp)
{
	struct device_board *device;
	unsigned long regval;

	for (device = boards; device; device = device->next) {
		if (MINOR(inode->i_rdev) == device->minor) {
			if (device->busy) {
				printk(KERN_INFO GSC_NAME "(%d): board already opened\n", device->minor);
				return (-EBUSY);
			}
			printk(KERN_INFO GSC_NAME "(%d): opening board\n", device->minor);
			if (request_irq(device->irqlevel, device_interrupt, SA_SHIRQ, GSC_NAME, device) < 0) {
				printk(KERN_ERR GSC_NAME "(%d): can not get interrupt %d\n", device->minor, device->irqlevel);
				device->busy = 0;
				return (-EBUSY);
			}
			if (!(device->intermediateBuffer = (u32 *)__get_dma_pages(GFP_KERNEL, DMA_ORDER))) {
				printk(KERN_ERR GSC_NAME "(%d): can not allocate DMA pages\n", device->minor);
				free_irq(device->irqlevel, device);
				device->busy = 0;
				return (-EBUSY);
			}
			device->intermediatePhysicalAddr = virt_to_phys((void *)device->intermediateBuffer);
#if DEBUG
			printk(KERN_INFO GSC_NAME "(%d): DMA virt=%lx, phys=0x%lx size: %d 0x%X\n", device->minor, (unsigned long)device->intermediateBuffer, (unsigned long)device->intermediatePhysicalAddr,(int)DMA_SIZE,(int)DMA_SIZE);
#endif
			device->busy = 1;
			device->dmaState = DMA_DISABLE;
			device->intermediateStart = 0;
			device->intermediateSamples = 0;
			device->dma_pending = FALSE;
			device->error = DEVICE_SUCCESS;
			device->signalno = (-1);
			device->signalev = NO_EVENT;
			device->init_pending = FALSE;
			device->autocal_pending = FALSE;
			device->chready_pending = FALSE;
			device->lo_hi_pending = FALSE;
			init_timer(&device->watchdog_timer);
			device->timeout_seconds=DEFAULT_TIMEOUT; //default
			device->watchdog_timer.function=timeout_handler;
			device->watchdog_timer.data = (unsigned long) device;
			device->bufferThreshold= 0xFFFE; // device default
			device->fillBuffer=FALSE;


#ifdef LINUX_2_4
			init_waitqueue_head(&device->ioctlwq);
			init_waitqueue_head(&device->readwq);
			init_waitqueue_head(&device->dmawq);
#else
			device->ioctlwq = NULL;
			device->readwq = NULL;
			device->dmawq = NULL;
#endif  
			fp->private_data = device;
			/* reset DMA engines */
#ifdef DEBUG
			printk(KERN_INFO GSC_NAME "(%d): resetting DMA engines\n", device->minor);
#endif
			regval = 0;
			regval &= STOP_DMA_CMD_0_MASK;
			regval &= STOP_DMA_CMD_1_MASK;
			writel(regval, DMACmdStatus(device));
			/* reset local IRQ flag */
#ifdef DEBUG
			printk(KERN_INFO GSC_NAME "(%d): disabling interrupts\n", device->minor);
#endif
			//`			writelocal(device,0, IntCtrlReg(device));
			//writel(0, (IntCntrlStat(device)));
#ifdef DEBUG
			printk(KERN_INFO GSC_NAME "(%d): open done\n", device->minor);
#endif
			MOD_INC_USE_COUNT;
#ifdef DEBUG
			regdump(device);
#endif
			return (0);
		}
	}
	printk(KERN_INFO GSC_NAME "(%d): can not find board\n", MINOR(inode->i_rdev));
	return (-ENODEV);
}

/************************************************************************/
/* close device                                                         */
/************************************************************************/
static int device_close(struct inode *inode, struct file *fp)
{
	struct device_board *device = (struct device_board *)fp->private_data;
	unsigned long regval;
	printk(KERN_INFO GSC_NAME "(%d): closing board\n", device->minor);
	/* disable local interrupts */
	regval = readl(IntCntrlStat(device));
	regval &= (~PCI_INT_ENABLE);
	writel(regval, (IntCntrlStat(device)));
	/* reset DMA engines */
	regval = 0;
	regval &= STOP_DMA_CMD_0_MASK;
	regval &= STOP_DMA_CMD_1_MASK;
	writel(regval, DMACmdStatus(device));
	/* reset local IRQ flag */
	//`	writelocal(device,0, IntCtrlReg(device));
	/* free resources */
	free_irq(device->irqlevel, device);
	free_pages((unsigned long)device->intermediateBuffer, DMA_ORDER);
	device->busy = 0;
	MOD_DEC_USE_COUNT;
	return (0);
}

/************************************************************************/
/* write operation: alway an error                                      */
/************************************************************************/
static int device_write(struct file *fp, const char *buf, size_t size, loff_t * lt)
{
	return (-EPERM);
}
/************************************************************************/
/* Spurious Interrupt handler                                           */
/************************************************************************/
#define CLEAR_SPURIOUS_PLX(regval,mask, desc)  { \
	if (regval & mask) \
{ \
	printk(KERN_ERR GSC_NAME "(%d): Unexpected interrupt %s %.8lX %.8X %.8X\n",device->minor,desc, regval,mask,~mask); \
	writel((regval & ~mask), IntCntrlStat(device)); \
} \
}
/*printk(KERN_INFO GSC_NAME "(%d): Unexpected interrupt %s %.8lX %.8X %.8X\n",device->minor,desc, regval,mask,~mask); \*/

/************************************************************************/
/* Interrupt handler                                                    */
/************************************************************************/
void device_interrupt(int irq, void *dev_id, struct pt_regs *regs)
{
	struct device_board *device = (struct device_board *)dev_id;
	unsigned long local_irq_reg;
	unsigned long plx_irq_reg;
	unsigned long dma_reg;
	int irq_type;
#if DEBUG
	printk(KERN_INFO GSC_NAME " Interrupt()!!\n");
#endif
	/* determine if the interrupt is from this board. */

	local_irq_reg = readlocal(device,INT_CTRL_REG);
	plx_irq_reg = readl(IntCntrlStat(device));

	if ((plx_irq_reg & IRQ_PCI_ENABLE) == 0) { /* not ours */
		//printk(KERN_INFO GSC_NAME "Interrupt not ours... %.8lx\n",plx_irq_reg);
#ifdef DEBUG
		int_other_count[device->board_index]++;
#endif
		return;
	}
	//printk(KERN_INFO GSC_NAME " Interrupt->ours! plx irq: %.8lX local irq: %.8lX plxdma stat: %.8X \n", plx_irq_reg,local_irq_reg,readl(DMACmdStatus(device)));
#if DEBUG
	writel(0x11111111, (Mailbox0(device)));
#endif
	plx_irq_reg &= ~PCI_INT_ENABLE;
	writel(plx_irq_reg, (IntCntrlStat(device)));

	/* check for DMA complete interrupt */
	if (device->dma_pending) {
		dma_reg = readl(DMACmdStatus(device));
		//printk(KERN_INFO GSC_NAME "DMA in progress. %.8lX\n",dma_reg);
		if (dma_reg & DMA0_DONE) {
			//writel(0xbbbb, (Mailbox0(device)));
			/* acq interrupt */
			writel(DMA0_CLR_INT, DMACmdStatus(device));
			/* wake up process blocked in 'read()' */
			device->dma_pending = FALSE;
			wake_up(&device->dmawq);
			//printk(KERN_INFO GSC_NAME "DMA end return plx irq %.8X local irq %.8X. plxdma stat%.8X.\n",readl(IntCntrlStat(device)),readlocal(device,INT_CTRL_REG),readl(DMACmdStatus(device)));
			//writel(0xdddd, (Mailbox0(device)));
			return;
		}
	}
	/*
	*	handle group 0 interrupts.
	*/
	//printk(KERN_INFO GSC_NAME "before Group 0 interrupt, local irq reg %.8lx\n",local_irq_reg);

	if (local_irq_reg&ICR_0_REQUEST)
	{
		irq_type=local_irq_reg&ICR_0_MASK;
		//writelocal(device,local_irq_reg & (~(IRQ_0_REQUEST|IRQ_0_MASK)), INT_CTRL_REG);
		writelocal(device,0, INT_CTRL_REG);
		//printk(KERN_INFO GSC_NAME "Group 0 interrupt. type: %x\n",irq_type);

		switch(irq_type) {
	case ICR_0_IDLE:
		//printk(KERN_INFO GSC_NAME ": group 0 IRQ_0_IDLE Interrupt %x\n",irq_type);
		if (device->init_pending) {
			device->init_pending = FALSE;
			wake_up(&device->ioctlwq);
		}
		break;
	case ICR_0_AUTOCAL_COMPLETE:
		//printk(KERN_INFO GSC_NAME ": group 0 IRQ_0_AUTOCAL_COMPLETE Interrupt %x\n",irq_type);
		if (device->autocal_pending) {
			device->autocal_pending = FALSE;
			wake_up(&device->ioctlwq);
		}
		break;
	case ICR_0_INPUT_SCAN_INIT:
		//printk(KERN_INFO GSC_NAME ": group 0 IRQ_0_INPUT_SCAN_INIT Interrupt %x\n",irq_type);
		if (device->init_pending) {
			device->init_pending = FALSE;
			wake_up(&device->ioctlwq);
		}
		break;
	case ICR_0_INPUT_SCAN_COMPLETE:
#ifdef DEBUG
		printk(KERN_INFO GSC_NAME "(%d): ICR_0_INPUT_SCAN_COMPLETE IRQ\n", device->minor);
#endif
		if (device->chready_pending) {
#ifdef DEBUG
			printk(KERN_INFO GSC_NAME "(%d): ICR_0_INPUT_SCAN_COMPLETE IRQ was waited for\n", device->minor);
#endif
			device->chready_pending = FALSE;
			wake_up(&device->dmawq);
		}
		break;
	default:
		printk(KERN_ERR GSC_NAME ": ERROR - unknown group 0 Interrupt\n");
		break;
		}
	}
	/*
	*	handle group 1 interrupts.
	*/
	if (local_irq_reg&ICR_1_REQUEST)
	{
		//		printk(KERN_INFO GSC_NAME "Group 1 interrupt.\n");
		//	writelocal(device,local_irq_reg & (~(IRQ_1_REQUEST|IRQ_1_MASK)), INT_CTRL_REG);
		writelocal(device,0, INT_CTRL_REG);
		irq_type=local_irq_reg&ICR_1_MASK;

		switch(irq_type) {
	case ICR_1_LO_HI:
		//		printk(KERN_INFO GSC_NAME ": IRQ_1_LO_HI 1 Interrupt %x\n",irq_type);
		if (device->lo_hi_pending) {
#ifdef DEBUG
			printk(KERN_INFO GSC_NAME "(%d): channels ready IRQ was waited for\n", device->minor);
#endif
			device->lo_hi_pending = FALSE;
			wake_up(&device->dmawq);
		} 
		break;
	case ICR_1_HI_LO:
		//		printk(KERN_INFO GSC_NAME ": IRQ_1_HI_LO 1 Interrupt %x\n",irq_type);
		break;
		}
	}
	/*
	*	check every possible interrupt source, to ensure that the
	* current interrupt is turned off, expected or not.
	*/
	/*
	printk(KERN_INFO GSC_NAME "spurious check.\n");

	// PLX interrupt-control registers.
	CLEAR_SPURIOUS_PLX(plx_irq_reg,IRQ_LOCAL_LSERR_ABORT, "IRQ_LOCAL_LSERR_ABORT*");
	CLEAR_SPURIOUS_PLX(plx_irq_reg, IRQ_LOCAL_LSERR_ABORT, "IRQ_LOCAL_LSERR_ABORT*");
	CLEAR_SPURIOUS_PLX(plx_irq_reg, IRQ_LOCAL_LSERR_OVERFLOW, "IRQ_LOCAL_LSERR_OVERFLOW*");
	CLEAR_SPURIOUS_PLX(plx_irq_reg, IRQ_GENERATE_SERR, "IRQ_GENERATE_SERR*");
	CLEAR_SPURIOUS_PLX(plx_irq_reg, IRQ_MAILBOX_ENABLE, "IRQ_MAILBOX_ENABLE*");
	//	CLEAR_SPURIOUS_PLX(plx_irq_reg, IRQ_PCI_ENABLE, "IRQ_PCI_ENABLE*");
	CLEAR_SPURIOUS_PLX(plx_irq_reg, IRQ_PCI_DOORBELL_ENABLE, "IRQ_PCI_DOORBELL_ENABLE*");
	CLEAR_SPURIOUS_PLX(plx_irq_reg, IRQ_ABORT_ENABLE, "IRQ_ABORT_ENABLE*");
	//	CLEAR_SPURIOUS_PLX(plx_irq_reg, IRQ_LOCAL_PCI_ENABLE, "IRQ_LOCAL_PCI_ENABLE*");
	//	CLEAR_SPURIOUS_PLX(plx_irq_reg, IRQ_LOCAL_ENABLE, "IRQ_LOCAL_ENABLE*");
	CLEAR_SPURIOUS_PLX(plx_irq_reg, IRQ_LOCAL_DOORBELL_ENABLE, "IRQ_LOCAL_DOORBELL_ENABLE*");
	CLEAR_SPURIOUS_PLX(plx_irq_reg, IRQ_DMA_1_ENABLE,"IRQ_DMA_1_ENABLE*");
	CLEAR_SPURIOUS_PLX(plx_irq_reg, IRQ_LOCAL_DOORBELL_ACTIVE,"IRQ_LOCAL_DOORBELL_ACTIVE*");
	CLEAR_SPURIOUS_PLX(plx_irq_reg, IRQ_DMA_1_ACTIVE,"IRQ_DMA_1_ACTIVE*");
	*/

	/* if something is pending, leave with the interrupt enabled. */

	if (device->dma_pending || device->autocal_pending || device->chready_pending || 
		device->lo_hi_pending || device->init_pending)
	{
		writel(readl(IntCntrlStat(device)) | IRQ_PCI_ENABLE | IRQ_LOCAL_PCI_ENABLE, IntCntrlStat(device));
	}

	//	printk(KERN_INFO GSC_NAME "IRQ end return plx irq %.8X local irq %.8X plxdma stat%.8X.\n",readl(IntCntrlStat(device)),readlocal(device, INT_CTRL_REG),readl(DMACmdStatus(device)));
	//writel(0xcccc, (Mailbox0(device)));

	return;
}

/************************************************************************/
/* writelocal                                                           */
/*                                                                      */
/* Write to the board local registers, whether the local registers      */
/* are in I/O space or memory space.                                    */
/*                                                                      */
/*                                                                      */
/************************************************************************/
void writelocal(struct device_board *device, unsigned value, unsigned address)
{
	if (device->PciBar[2].IsIoMapped) /* i/o space */
	{
		//printk(KERN_INFO GSC_NAME "writelocal I/O space addr: %.8X value: %.8X.\n",address,value);
		outl(value, (unsigned) device->local_addr+address*4);
	}
	else
	{
		writel(value, device->local_addr+address);
	}
};

/************************************************************************/
/* readlocal                                                            */
/*                                                                      */
/* Read from the board local registers, whether the local registers     */
/* are in I/O space or memory space.                                    */
/*                                                                      */
/*                                                                      */
/************************************************************************/
unsigned readlocal(struct device_board *device, unsigned address)
{
	if (device->PciBar[2].IsIoMapped) /* i/o space */
	{
		//		printk(KERN_INFO GSC_NAME "readlocal I/O space addr: %.8X \n",address);
		return inl((unsigned)device->local_addr+address*4);
	}
	else
	{
		return readl(device->local_addr+address);
	}
	return 0;
};
