/***
*** main.c  
***
***  General description of this file:
***     Device driver source code for General Standards 16AI64
***     family of analog input boards. This file is part of the Linux
***     driver source distribution for this board.
***     
***
***  Copyrights (c):
***     General Standards Corporation (GSC), 2004-2005
***
***  Author:
***     Evan Hillman (evan@generalstandards.com)
***
***  Support:
***     Primary support for this driver is provided by GSC. 
***
***/

//////////////////////////////////////////////////////////////////////////
// set the following flag to trace interrupt debug messages.
#ifdef DEBUG
#define TRACE_LOCAL TRUE
#endif

#include <linux/module.h>
#include <linux/fs.h>
#include <linux/vmalloc.h>
#include <linux/pci.h>
#include <linux/mm.h>
#include <linux/proc_fs.h>
#include <linux/sched.h>
#include <linux/interrupt.h>
#include <linux/timer.h>
#include <linux/ioctl.h>

#include <asm/io.h>
#include <asm/system.h>
#include <asm/uaccess.h>

#include "sysdep.h"
#include "gsc16ai64_ioctl.h"
#include "internals.h"
#include "plx_regs.h"

MODULE_LICENSE("GPL");

struct board_entry boards_supported[] = {
    /*    subvendor    subsystem name       index ID */
    {GSC_SUBVENDOR,0x2406,"PMC-12AI64",gsc12ai64}, /* PMC-12AI64 */
    {GSC_SUBVENDOR,0x2407,"PMC-16AI64",gsc16ai64}, /* PMC-16AI64 */
    {GSC_SUBVENDOR,0x2868,"PMC-16AI64SS",gsc16ai64ss}, /** PCM-16AI64SS */
    {0,0,"null"}               /* null terminated list.... */
};

/* Driver Version */
static const char device_version[] = "$Date: " __DATE__ " " __TIME__ " $";

/* module load/unload functions */
int init_module(void);
void cleanup_module(void);

/* local functions */
int device_write(struct file *, const char *, size_t, loff_t *);
int device_read(struct file *, char *, size_t, loff_t *);
int device_open(struct inode *, struct file *);
int device_close(struct inode *, struct file *);

int device_read_buffer(struct device_board *device, char *buf, int nsamp, int noblock);

int device_read_dma(struct device_board *device, int nsamples);

#ifdef DEBUG
    u32 int_other_count[MAX_BOARDS];
    u32 int_count[MAX_BOARDS];
    u32 dma_count[MAX_BOARDS];
    u32 channel_irq[MAX_BOARDS];
    u32 channel_expected[MAX_BOARDS];
    int board_type[MAX_BOARDS];
    void * context[MAX_BOARDS];
#endif

//#ifdef DEBUG
/************************************************************************/
/* regdump - debug only                                                 */
/************************************************************************/
void regdump(struct device_board *device,char * label)
{
    int q;

    for (q=0;q<=12;q++)
    {
        msg("(%d - %s): register %.2d (addr %.2X) value: %.8x\n", device->minor, label, q, q*4,readlocal(device,q));
    }
    msg("(%d - %s)                 IntCntrlStat: %.8X \n", device->minor, label,readl(IntCntrlStat(device)));
    msg("(%d - %s)                   DMA status: %.8X \n", device->minor, label,readl(DMACmdStatus(device)));
    msg("(%d - %s)                     DMA Mode: %.8X \n", device->minor, label,readl(DMAMode0(device)));
    for (q=0;q<=12;q++)
    {
        printk("%.8X ",device->intermediateBuffer[q]);
    }
    msg("\n----------------------------------\n");

}
//#endif

/************************************************************************/
/* timeout_handler                                                      */
/************************************************************************/
void timeout_handler(unsigned long ptr)
{
    struct device_board *device = (struct device_board *) ptr;
    int i;

    mark(0xAaaaaaaa);
    errmsg("(%d): ERROR - device timeout - Debug state: %d\n", device->minor,device->debug_state);
    regdump(device,"timeout " );
    device ->timeout=TRUE;

    // wake up whatever was waiting.
    {
        int fifosamples,threshold;
        fifosamples = readlocal(device,BUFFER_SIZE_REG);
        threshold = readlocal(device, INPUT_BUFF_CTRL_REG)&device->hwBufferSize;
        errmsg("(%d): timeout -> fifo samples %.8X threshold %.8X\n", device->minor, fifosamples, threshold);
    }
    errmsg("timeout -> plx int %.8X board control %.8X DMA status %.8X buffer size: %.8X \n",
        readl(IntCntrlStat(device)), readlocal(device,BOARD_CTRL_REG),
        readl(DMACmdStatus(device)),readlocal(device,BUFFER_SIZE_REG));
    errmsg("DMA Mode %.8X\n",readl(DMAMode0(device)));
    errmsg("Firm rev %.8X\n",readlocal(device,0x3c/4));

    if (atomic_read(&device->irq_event_pending[EVENT_DMA_PENDING]))
    {
        writel((DMA0_CLR_INT|DMA0_ABORT), DMACmdStatus(device)); 
    }

    for(i=0;i<EVENT_LAST_IRQ_EVENT;i++)
    {
        if (atomic_read(&device->irq_event_pending[i]))
        {
            errmsg("(%d): ERROR - event %d timeout\n", device->minor, i);
            atomic_set(&device->irq_event_pending[i],FALSE);
            wake_up(device->event_queue_waiting[i]);
        }
    }

    errmsg("(%d): End timeout handler\n", device->minor);
}

/*************************************************/
#define CLEAR_SPURIOUS_PLX(regval,mask, desc)  { \
    if (regval & mask) \
{ \
    writel((regval & ~mask), IntCntrlStat(device)); \
} \
}

/************************************************************************/
/* Interrupt handler                                                    */
/************************************************************************/
#if LINUX_VERSION_CODE >= KERNEL_VERSION(2,6,0)
irqreturn_t device_interrupt(int irq, void *dev_id, struct pt_regs *regs)
#else
void device_interrupt(int irq, void *dev_id, struct pt_regs *regs)
#endif
{
    struct device_board *device = (struct device_board *)dev_id;
    u32 plxreg;
    u32 localreg;
    u32 dmareg;
    int handled;
    int local_irq;

    handled = FALSE;

    //msg(" Interrupt()\n");
    mark(0x11111130);

    // determine if the interrupt is enabled.

    plxreg = readl(IntCntrlStat(device));
    if (((plxreg & IRQ_PCI_ENABLE)==0) ||
        (((plxreg & IRQ_DMA_0_ACTIVE)==0)&&((plxreg & IRQ_LOCAL_ACTIVE))==0)) 
    { // not ours 
        //msg("Interrupt not ours... %.8X\n",plxreg);
#ifdef TRACE_LOCAL
        mark(0x12121212);
        int_other_count[device->board_index]++;
        //if (int_other_count[device->board_index]>10)
        //{
        //    mesg("Interrupt not ours... killing...\n");
        //    DisableIrqPlx(device);
        //}
#endif
#if LINUX_VERSION_CODE >= KERNEL_VERSION(2,6,0)
        return IRQ_NONE;
#else
        return;             
#endif
    }

    // disable the PCI interrupt.

    DisableIrqPlx(device);

    // deal with local interrupt sources.

    localreg = readlocal(device,BOARD_CTRL_REG);
#ifdef TRACE_LOCAL
    mark(0x13131313);
    //dmesg("(%d): our Interrupt! plx: %.8X BCR: %.8X\n", device->minor,plxreg,localreg);
    int_other_count[device->board_index]=0;
#endif

    // check for DMA complete interrupt
    if (atomic_read(&device->irq_event_pending[EVENT_DMA_PENDING])) {
        //dmsg("(%d): IRQ DMA in progress...\n",device->minor);
        dmareg = readl(DMACmdStatus(device));
        if (dmareg & DMA0_DONE) {
#ifdef TRACE_LOCAL
            //demsg("(%d): DMA in progress: %d PLX int/ctrl %.8X BC %.8X DMA status %.8X DMA mode %.8X\n", device->minor,atomic_read(&device->irq_event_pending[EVENT_DMA_PENDING]), readl(IntCntrlStat(device)), readlocal(device,BOARD_CTRL_REG),readl(DMACmdStatus(device)),readl(DMAMode0(device)));
            dma_count[device->board_index]++;
/*
            {
                int q;
                for(q=0;q<10;q++)
                {
                    printk("%.8X ", device->dma_data[RX].intermediateBuffer[q]);
                }
                printk("\n");
            }
            */
#endif
            handled = TRUE;
            // ack interrupt 
            writel((dmareg | DMA0_CLR_INT), DMACmdStatus(device));
            writel((DMA0_CLR_INT), DMACmdStatus(device)); 

            // wake up process blocked in 'read()'
            atomic_set(&device->irq_event_pending[EVENT_DMA_PENDING],FALSE);
            wake_up(device->event_queue_waiting[EVENT_DMA_PENDING]);
        }
    }
    else    // check for and clear spurious DMA interrupts...
    {
        plxreg = readl(DMACmdStatus(device));
        if (plxreg & DMA0_ENABLE) {
            // wipe out the DMA1 bits while here, better safe than sorry.
            errmsg("(%d): Unexpected interrupt DMA0_ENABLE\n",device->minor); 
            writel((DMA0_CLR_INT|CH1_DMA_CLEAR_IRQ_MASK), DMACmdStatus(device)); 
        }
    }

    // check for local interrupt

    localreg = readlocal(device,INTERRUPT_CONTROL_REG);
    //DisableIrqLocal(device);
    //dmesg(KERN_INFO GSC_NAME "(%d): plx: %.8X interrupt control %.8X \n", device->minor,plxreg,localreg);

    // check group 0 sources.

    if (localreg & ICR_GROUP_0_REQUEST) 
    {
#ifdef TRACE_LOCAL
        int_count[device->board_index]++;
        //dmesg("(%d): Local group 0!\n",device->minor);
#endif

        // see which kind of event is this.  The hardware is assumed to only have one
        // interrupt active at a time.

        local_irq=(localreg>> ICR_GROUP_0_SHIFT) & ICR_GROUP_0_MASK; 
        DisableIrqLocal(device,GROUP_0,0);

        msg(KERN_INFO GSC_NAME "(%d): irq %d \n", device->minor,local_irq);
        if (atomic_read(&device->irq_event_pending[GROUP_0_EVENT+local_irq])) {
            mark(0x14141414);
            atomic_set(&device->irq_event_pending[GROUP_0_EVENT+local_irq],FALSE);
            wake_up(device->event_queue_waiting[GROUP_0_EVENT+local_irq]);
            handled = TRUE;
        }
        else
        {
            mark(0x15151515);
            // msg("(%d): irq %s %p\n", device->minor,event_name[local_irq], device);
        }
    }

    // check group 1 sources.

    if (localreg & ICR_GROUP_1_REQUEST) 
    {
#ifdef TRACE_LOCAL
        int_count[device->board_index]++;
        //msg("(%d): Local group 1!\n",device->minor);
#endif

        // see which kind of event is this.  The hardware is assumed to only have one
        // interrupt active at a time.

        local_irq=(localreg>>ICR_GROUP_1_SHIFT) & ICR_GROUP_1_MASK ; 
        DisableIrqLocal(device,GROUP_1,0);

        //msg("(%d): irq %d\n", device->minor,local_irq);
        if (atomic_read(&device->irq_event_pending[GROUP_1_EVENT+local_irq])) {
            mark(0x14141414);
            atomic_set(&device->irq_event_pending[GROUP_1_EVENT+local_irq],FALSE);
            wake_up(device->event_queue_waiting[GROUP_1_EVENT+local_irq]);
            handled = TRUE;
        }
        else
        {
            mark(0x15151515);
            //msg("(%d): irq %s %p\n", device->minor,event_name[local_irq], device);
        }
    }

    // clear any of the active interrupts.

    localreg = readlocal(device,INTERRUPT_CONTROL_REG);
    localreg &= ~(ICR_GROUP_0_REQUEST|ICR_GROUP_1_REQUEST);
    writelocal(device,localreg,INTERRUPT_CONTROL_REG);

    /*
    *   check every possible interrupt source, to ensure that the
    * current interrupt is turned off, expected or not.
    * commented out are at times expected.
    */

    // PLX interrupt-control registers.
    CLEAR_SPURIOUS_PLX(plxreg,IRQ_LOCAL_LSERR_ABORT, "IRQ_LOCAL_LSERR_ABORT*");
    CLEAR_SPURIOUS_PLX(plxreg, IRQ_LOCAL_LSERR_ABORT, "IRQ_LOCAL_LSERR_ABORT*");
    CLEAR_SPURIOUS_PLX(plxreg, IRQ_LOCAL_LSERR_OVERFLOW, "IRQ_LOCAL_LSERR_OVERFLOW*");
    CLEAR_SPURIOUS_PLX(plxreg, IRQ_GENERATE_SERR, "IRQ_GENERATE_SERR*");
    CLEAR_SPURIOUS_PLX(plxreg, IRQ_MAILBOX_ENABLE, "IRQ_MAILBOX_ENABLE*");
    //  CLEAR_SPURIOUS_PLX(plxreg, IRQ_PCI_ENABLE, "IRQ_PCI_ENABLE*");
    CLEAR_SPURIOUS_PLX(plxreg, IRQ_PCI_DOORBELL_ENABLE, "IRQ_PCI_DOORBELL_ENABLE*");
    CLEAR_SPURIOUS_PLX(plxreg, IRQ_ABORT_ENABLE, "IRQ_ABORT_ENABLE*");
    //  CLEAR_SPURIOUS_PLX(plxreg, IRQ_LOCAL_PCI_ENABLE, "IRQ_LOCAL_PCI_ENABLE*");
    //  CLEAR_SPURIOUS_PLX(plxreg, IRQ_LOCAL_ENABLE, "IRQ_LOCAL_ENABLE*");
    CLEAR_SPURIOUS_PLX(plxreg, IRQ_LOCAL_DOORBELL_ENABLE, "IRQ_LOCAL_DOORBELL_ENABLE*");
    CLEAR_SPURIOUS_PLX(plxreg, IRQ_DMA_1_ENABLE,"IRQ_DMA_1_ENABLE*");
    CLEAR_SPURIOUS_PLX(plxreg, IRQ_LOCAL_DOORBELL_ACTIVE,"IRQ_LOCAL_DOORBELL_ACTIVE*");
    CLEAR_SPURIOUS_PLX(plxreg, IRQ_DMA_1_ACTIVE,"IRQ_DMA_1_ACTIVE*");

    //msg(" (%d): end IRQ: %d PLX int/ctrl %.8X BC %.8X DMA status %.8X DMA mode %.8X\n", device->minor,device->dma_pending, readl(IntCntrlStat(device)), readlocal(device,BOARD_CTRL_REG),readl(DMACmdStatus(device)),readl(DMAMode0(device)));

    if(!handled)
    {
        mark(0x16161616);
        errmsg("(%d): >>>>>>>>>>>>>Unhandled interrupt - disabling\n", device->minor);
        DisableIrqLocalAll(device);
        DisableIrqPlx(device);
        timeout_handler((unsigned long)device);

    }
    mark(0x22222222);
    //msg("(%d): End IRQ handler\n", device->minor);
#if LINUX_VERSION_CODE >= KERNEL_VERSION(2,6,0)
    return IRQ_HANDLED;
#else
    return;             
#endif
}
