/***
*** read.c  
***
***  General description of this file:
***     Device driver source code for General Standards 16AI64
***     family of analog input boards. This file is part of the Linux
***     driver source distribution for this board.
***     
***
***  Copyrights (c):
***     General Standards Corporation (GSC), 2004-2005
***
***  Author:
***     Evan Hillman (evan@generalstandards.com)
***
***  Support:
***     Primary support for this driver is provided by GSC. 
***
***/
#ifndef __KERNEL__
#define __KERNEL__
#endif

#ifndef MODULE
#define MODULE
#endif
//////////////////////////////////////////////////////////////////////////
// set the following flag to trace interrupt debug messages.
#ifdef DEBUG
#define TRACE_LOCAL TRUE
#endif

#include <linux/module.h>
#include <linux/fs.h>
#include <linux/vmalloc.h>
#include <linux/pci.h>
#include <linux/mm.h>
#include <linux/proc_fs.h>
#include <linux/sched.h>
#include <linux/interrupt.h>
#include <linux/timer.h>
#include <linux/ioctl.h>

#include <asm/io.h>
#include <asm/system.h>
#include <asm/uaccess.h>

#include "sysdep.h"
#include "gsc16AI64_ioctl.h"
#include "internals.h"
#include "plx_regs.h"

static int FillIntermediateBuffer(struct device_board *device, u32 *buf, unsigned long nRequested, int noblock);
static int device_read_dma(struct device_board *device, unsigned long nRequested,int DMAMode);

/************************************************************************/
/* TransferToUser - be sure not to copy more than there is space in the */
/* user buffer.  Returns bytes transferred.                             */
/************************************************************************/
long TransferToUser(struct device_board *device, unsigned long nRequested)
{
	int bytesRequested;
	unsigned long ncopied;

	// make sure there is space for the data.
	nRequested = min(device->userRemaining, nRequested);

	// make sure there is data
	nRequested = min(device->intermediateSamples, nRequested);

	// don't transfer more than is in the intermediate buffer.
	nRequested = min(DMA_SIZE, nRequested);

	// convert to bytes.
	bytesRequested = nRequested * sizeof(u32);

	msg("(%d): TransferToUser before userRemain %ld intermediateSamples %ld nRequested %ld\n",device->minor, device->userRemaining,device->intermediateSamples, nRequested);
	msg("(%d): TransferToUser before userOffset %ld intermediateSamples %ldnRequested %ld\n",device->minor, (unsigned long)device->userOffset, device->intermediateSamples, nRequested);
	msg("(%d): src base %p starting src %p dst base %p starting dst %p\n",device->minor, device->userBuffer,((char*)device->userBuffer)+device->userOffset*sizeof(u32), device->intermediateBuffer, &device->intermediateBuffer[device->intermediateStart]);
	msg("(%d): intermediate start %d\n", device->minor,device->intermediateStart);

	ncopied = __copy_to_user(((char*)device->userBuffer)+device->userOffset*sizeof(u32), 
		           &device->intermediateBuffer[device->intermediateStart], bytesRequested);

	device->intermediateStart += nRequested;
	device->intermediateSamples -= nRequested;
	device->userOffset += nRequested;
	device->userRemaining-= nRequested;
	msg("(%d): TransferToUser after userRemain %ld intermediateSamples %ld nRequested %ld\n",device->minor, device->userRemaining,device->intermediateSamples, nRequested);
	msg("(%d): TransferToUser after userOffset %ld intermediateSamples %ld nRequested %ld\n",device->minor, (unsigned long)device->userOffset, device->intermediateSamples, nRequested);
	mark(0x11111100);

	return bytesRequested;
}

/************************************************************************/
/* Transfer data from the intermediate buffer to the user buffer.       */
/* returns bytes transferred                                            */
/************************************************************************/
int ProcessRead(struct device_board *device)
{
	unsigned long nRequested;
	int retval;
	long bytesCopied;

	nRequested=device->userRemaining;

	bytesCopied=0;

	// start by copying what is aleady in the intermediate buffer.

	msg("(%d): ProcessRead - residual samples: %ld\n", device->minor,device->intermediateSamples);
	msg("(%d): intermediate start %d\n", device->minor,device->intermediateStart);

	// give either what the user requests or what is left in the buffer, whichever is less.
	if (device->intermediateSamples > 0) {

		msg("(%d): ProcessRead - local data copy...\n", device->minor);
		retval = TransferToUser(device, nRequested);

		// if the user buffer is now full, return.
		if (device->userRemaining==0){ // done
			mark(0x11111101);

			return(retval);
		}

		// if the user does not ask for full buffers, return
		if (!device->fillBuffer) {
			mark(0x11111102);
			return(retval);
		}

	} // if (device->intermediateSamples > 0) 

	// need to get a batch of data from the board.

	// read at most an intermediate buffer full at a time
	if (nRequested > DMA_SAMPLES) {
		nRequested = DMA_SAMPLES;
	}

	do {  // do it over and over if fillBuffer flag set, once otherwise
		msg("(%d): ProcessRead - looping...remaining: %ld fillbuffer %d user offset %d %x\n", device->minor,device->userRemaining,device->fillBuffer,device->userOffset,device->userOffset);
			retval = FillIntermediateBuffer(device, device->userBuffer+device->userOffset, 
				nRequested, (device->fp->f_flags & O_NONBLOCK));
			if (retval<0){ // returned error
				mark(0x11111103);
				return retval;
			}
			if (retval==0) { // got no data
				mark(0x11111104);
				return (-EAGAIN);
			}
			bytesCopied += TransferToUser(device, retval);
	} while((device->userRemaining>0) && device->fillBuffer);
	
	msg("(%d): ProcessRead - returning: %ld bytes %ld samples\n", device->minor, bytesCopied, bytesCopied/sizeof(u32));
	mark(0x11111105);
	return (bytesCopied);
}

/************************************************************************/
/* read operation: either polled or uses PLX DMA on CH0                 */
/************************************************************************/
int device_read(struct file *fp, char *buf, size_t size, loff_t * lt)
{
	struct device_board *device = (struct device_board *)fp->private_data;
	unsigned long nRequested;
	unsigned long fifosamples,regval;
#ifdef DEBUG
//	memset(buf,0x55, size);
#endif

	device->fp=fp;
	nRequested = size / sizeof(u32);
#ifdef TRACE_LOCAL
    {
        static int nread;
        msg("(%d): device_read() data requested: %ld %lx -------------------%d\n", device->minor,nRequested,nRequested,nread);
        nread++;
    }
#endif

	// verify parameters 
	if (nRequested <= 0) {
		msg("(%d): device_read - zero sample return\n", device->minor);
		mark(0x11111106);
		return (0);
	}

	if (!access_ok(VERIFY_WRITE, (unsigned long)buf, size)) {
		errmsg("(%d): device_read - access not OK\n", device->minor);
		mark(0x11111107);
		return (-EFAULT);
	}

	if (device->detectOverflow) {
		fifosamples = readlocal(device,BUFFER_SIZE_REG);
		if (fifosamples>=device->hwBufferSize-1){ // overflow

			// clear the buffer.

			regval = readlocal(device,INPUT_BUFF_CTRL_REG);
			regval |= IDC_INPUT_CLEAR_BUFFER;
			writelocal(device,regval, INPUT_BUFF_CTRL_REG);
			device->error = DEVICE_BUFFER_OVERFLOW;
			mark(0x11111108);
			return (-EIO);
		}
	}
	device->userOffset=0; // start at the beginning of the user buffer.
	device->userRemaining=nRequested;
	device->intermediateStart=0;
	device->userBuffer=(u32*)buf;
	mark(0x11111109);
	return ProcessRead(device);
}

/************************************************************************/
/*  FillIntermediateBuffer                                              */
/*                                                                      */
/* Buffered I/O read method                                             */
/* The method first attempts to fill the DMA buffer to a "reasonable"   */
/* level                                                                */
/************************************************************************/
static int FillIntermediateBuffer(struct device_board *device, u32 *buf, unsigned long nRequested, int noblock)
{
	int fifosamples, i, threshold;
	unsigned long nToTransfer;

	nToTransfer = min(nRequested,DMA_SAMPLES);
	fifosamples=0;
	device->intermediateStart=0;
	device->intermediateSamples=0;

	msg("(%d): FillIntermediateBuffer() nRequested %ld DMA_SAMPLES %ld\n", device->minor,nRequested,DMA_SAMPLES);
	switch(device->dmaState) {
		// for PIO and regular DMA, wait for sufficient samples
		case DMA_DISABLE: // PIO
		case DMA_ENABLE:

			// enable the interrupt now in case the interrupt happens during the moment of indecision

			atomic_set(&device->irq_event_pending[EVENT_BUFFER_THRES_LOW_TO_HIGH],TRUE);
			EnableIrqLocal(device, GROUP_1, ICR_GROUP_1_LO_HI);

			// check how much data there is in the FIFO 
			fifosamples = readlocal(device,BUFFER_SIZE_REG);
			threshold = readlocal(device, INPUT_BUFF_CTRL_REG)&device->hwBufferSize;
		    msg("(%d): FillIntermediateBuffer first Current fifo samples 0x%.8X threshold 0x%.8X\n", device->minor, fifosamples, threshold);
			if (fifosamples < threshold) // need more
			{
				// first of all, check for non-blocking I/O.
				if (noblock) {
					msg("(%d): FillIntermediateBuffer - return error noblock\n", device->minor);
					device->error = DEVICE_SUCCESS;
					mark(0x11111110);
					return (-EAGAIN);
				}

				// see if there is already more data in the input data than the threshold.
				msg("(%d): FillIntermediateBuffer DMA/PIO wait for more data...\n", device->minor);

				EnableIrqPlxLocal(device);

				device->timeout=FALSE;
				//device->lo_hi_pending=TRUE;
				device->watchdog_timer.expires=jiffies+device->timeout_seconds*HZ;
				add_timer(&device->watchdog_timer);

				mark(0x11111141);
                if (wait_event_interruptible(*device->event_queue_waiting[EVENT_BUFFER_THRES_LOW_TO_HIGH],(!atomic_read(&device->irq_event_pending[EVENT_BUFFER_THRES_LOW_TO_HIGH]))))
				{
					errmsg("(%d): Signal termination...\n", device->minor);
					del_timer_sync(&device->watchdog_timer);
					DisableIrqLocalAll(device);
					DisableIrqPlx(device);
					mark(0x11111191);
					return (-ERESTARTSYS); 
				}
				if (signal_pending(current)) {
					errmsg("(%d): Signal termination...\n", device->minor);
					del_timer_sync(&device->watchdog_timer);
					DisableIrqLocalAll(device);
					DisableIrqPlx(device);
					mark(0x11111111);
					return (-ERESTARTSYS); 
				}
				mark(0x55555555);
				msg("(%d): FillIntermediateBuffer Next fifo samples %.8X threshold %.8X\n", device->minor, readlocal(device,BUFFER_SIZE_REG), readlocal(device, INPUT_BUFF_CTRL_REG)&device->hwBufferSize);
				if (device->timeout)
				{
					errmsg("(%d): FillIntermediateBuffer - timeout...\n", device->minor);


					//msg(" (%d): In progress: %d PLX int/ctrl %.8X BC %.8X DMA %.8X\n", device->minor,device->dma_pending, readl(IntCntrlStat(device)), readlocal(device,BOARD_CTRL_REG),readl(DMACmdStatus(device)));
					device->error = DEVICE_PIO_TIMEOUT;
					mark(0x11111112);
					return (-EIO);
				}
				else
					del_timer_sync(&device->watchdog_timer);

				if (signal_pending(current)) {
					errmsg("(%d): Signal termination...\n", device->minor);
					DisableIrqLocalAll(device);
					DisableIrqPlx(device);
					mark(0x11111113);
					return (-ERESTARTSYS); 
				}
				msg("(%d): FillIntermediateBuffer second Current fifo samples %d\n", device->minor, fifosamples);
				if (device->timeout) {
					msg("(%d): FillIntermediateBuffer - timeout error\n", device->minor);
					device->error = DEVICE_PIO_TIMEOUT;
					mark(0x11111114);
					return (-EIO);
				}
				fifosamples = readlocal(device,BUFFER_SIZE_REG);  // update samples available
				msg("(%d): after wait for data Updated buffer count:  %d \n", device->minor, fifosamples);
#if TRACE_LOCAL
				memset(device->intermediateBuffer,0xaa, DMA_SIZE);
#endif
				// if still no data present, something is wrong.

				if (fifosamples<=0) {
					errmsg("(%d): FillIntermediateBuffer - no data error\n", device->minor);
					mark(0x11111115);
					return (-EIO);
				}
			}
			else // don't need to wait for data...
			{
				DisableIrqLocalAll(device);
				atomic_set(&device->irq_event_pending[EVENT_BUFFER_THRES_LOW_TO_HIGH],FALSE);
			}

			break;

			nToTransfer=min(nToTransfer,(unsigned long)fifosamples);

			// for demand mode, wait for the DMA done interrupt.
		case DMA_DEMAND_MODE:
#ifdef TRACE_LOCAL
			msg("(%d): FillIntermediateBuffer Demand mode start... %ld\n", device->minor,nRequested);
			//				{
			//					int i;
			//					for(i=0;i<10;i++)
			//					{
			//						msg("(%d): %lX\n", device->minor, device->intermediateBuffer[i]);
			//					}
			//					i=0;
			//				}
#endif

			if (device_read_dma(device, nToTransfer, DMA_DEMAND_MODE)!=0)
			{
#ifdef TRACE_LOCAL
				msg("(%d): FillIntermediateBuffer - demand mode device_read_dma failure\n", device->minor);
				//					{
				//						int i;
				//						for(i=0;i<10;i++)
				//						{
				//							msg("(%d): %lX\n", device->minor, device->intermediateBuffer[i]);
				//						}
				//						while (device->intermediateBuffer[i]!=0) {
				//							i++;
				//						}
				//						msg("(%d): got samples:%d\n", device->minor,i);
				//					}
#endif
				mark(0x11111116);
				return (-EAGAIN);
			}
			else
			{
				msg("(%d): FillIntermediateBuffer - demand mode success\n", device->minor);
			};
			break;

		default: // invalid state
			errmsg("(%d): FillIntermediateBuffer - invalid state failure\n", device->minor);
			mark(0x11111117);
			return (-EIO);
			break;
	} // switch

	// move the data from the hardware buffer to the intermediate buffer by the user's chosen method.

	switch(device->dmaState) {
		case DMA_DISABLE: // PIO
			msg("(%d): PIO starting addr %p fifo samples %d to transfer %ld 0x%lX\n", device->minor,device->intermediateBuffer,fifosamples,nToTransfer,nToTransfer);
			// don't copy more than the buffer can hold.

			for (i=0; i< nToTransfer; i++) { // PIO read
				device->intermediateBuffer[i] = readlocal(device,INPUT_DATA_BUFF_REG);
#ifdef DEBUG
				if ((device->intermediateBuffer[i] & 0xff) == 0x20)
				{
					errmsg("Data error! %X\n",device->intermediateBuffer[i]);
				}
#endif
			}
			device->intermediateSamples=i;
			msg("(%d): PIO after addr %p fifo samples 0x%X moved %X\n", device->minor,&device->intermediateBuffer[i],fifosamples,i);
			break;

		case DMA_ENABLE:
			if (device_read_dma(device, nToTransfer,DMA_ENABLE)!=0)
			{
				msg("(%d): FillIntermediateBuffer - device_read_dma failure\n", device->minor);
				mark(0x11111118);
				return (-EAGAIN);
			};
			break;

		case DMA_DEMAND_MODE:
			// already in the intermediate buffer.
			break;

		default: // invalid state
			mark(0x11111119);
			return (-EIO);
			break;
	}

		msg("(%d): FillIntermediateBuffer after xfer fifo samples %d transferred: %ld\n", device->minor, readlocal(device,BUFFER_SIZE_REG),nToTransfer);

	mark(0x11111120);
	return nToTransfer;

} // FillIntermediateBuffer

/************************************************************************/
/* device_read_dma                                                      */
/************************************************************************/
/* fill the buffer using the DMA method */
/* by the time this is called, we know that we have enough samples in the FIFO */
static int device_read_dma(struct device_board *device, unsigned long nRequested, int DMAMode)
{
	unsigned long regval;
	msg("(%d): device_read_dma\n", device->minor);
	//regdump(device);

    atomic_set(&device->irq_event_pending[EVENT_DMA_PENDING],TRUE);
    EnableIrqPlxDMA(device);

	if (DMAMode==DMA_ENABLE) 
	{
		writel(NON_DEMAND_DMA_MODE, DMAMode0(device));
	}
	else
	{
		writel(DEMAND_DMA_MODE, DMAMode0(device));
	}
	writel(device->intermediatePhysicalAddr, DMAPCIAddr0(device));
	writel((INPUT_DATA_BUFF_REG << 2), DMALocalAddr0(device));
	writel((nRequested << 2), DMAByteCnt0(device));
	writel(0xA, DMADescrPtr0(device));
	writel(0, DMAArbitr(device));
	writel(0, DMAThreshold(device));
	regval = readl(DMACmdStatus(device));
	regval &= STOP_DMA_CMD_0_MASK;
	writel(regval, DMACmdStatus(device));
	// OK, here we go! 
	regval = readl(DMACmdStatus(device));
	regval |= SETUP_DMA_CMD_0;
	writel(regval, DMACmdStatus(device));
	regval |= STARTUP_DMA_CMD_0;
	mark(0xeeeeeeee);
	writel(regval, DMACmdStatus(device));

	msg("(%d): device_read_dma - about to sleep...\n", device->minor);

	device->timeout=FALSE;
	device->watchdog_timer.expires=jiffies+device->timeout_seconds*HZ;
	add_timer(&device->watchdog_timer);
    if (wait_event_interruptible(*device->event_queue_waiting[EVENT_DMA_PENDING],(!atomic_read(&device->irq_event_pending[EVENT_DMA_PENDING])))){
	errmsg("(%d): Signal termination...\n", device->minor);
	del_timer_sync(&device->watchdog_timer);
	DisableIrqLocalAll(device);
	DisableIrqPlx(device);
	mark(0x11111171);
	return (-ERESTARTSYS); 
	}

	if (signal_pending(current)) {
		msg("(%d): Signal termination...\n", device->minor);
		del_timer_sync(&device->watchdog_timer);
		DisableIrqLocalAll(device);
		DisableIrqPlx(device);
		mark(0x11111121);
		return (-ERESTARTSYS); 
	}

	if (device->timeout)
	{
		errmsg("(%d): device_read_dma - timeout...\n", device->minor);
//		msg(" (%d): In progress: %d PLX int/ctrl %.8X BC %.8X DMA %.8X\n", device->minor,device->dma_pending, readl(IntCntrlStat(device)), readlocal(device,BOARD_CTRL_REG),readl(DMACmdStatus(device)));
		mark(0x11111122);
		return (-EAGAIN);
	}
	else
	{
		msg(KERN_INFO GSC_NAME "(%d): device_read_dma - success...\n", device->minor);
		del_timer_sync(&device->watchdog_timer);
	}
	device->intermediateSamples=nRequested;
	mark(0x11111123);
	return 0;
}

